@@ -1,7 +1,6 @@
 #!/usr/bin/python3 -B
 import argparse
 import json
-import logging
 import os
 import shutil
 import subprocess
@@ -22,7 +21,6 @@ from typing import (
 )
 
 from debputy import DEBPUTY_ROOT_DIR, DEBPUTY_PLUGIN_ROOT_DIR
-from debputy._deb_options_profiles import DebBuildOptionsAndProfiles
 from debputy.analysis import REFERENCE_DATA_TABLE
 from debputy.analysis.debian_dir import scan_debian_dir
 from debputy.build_support import perform_clean, perform_builds
@@ -32,7 +30,8 @@ from debputy.commands.debputy_cmd.context import (
     ROOT_COMMAND,
     CommandArg,
 )
-from debputy.commands.debputy_cmd.output import _stream_to_pager
+from debputy.commands.debputy_cmd.output import _stream_to_pager, _output_styling
+from debputy.commands.debputy_cmd.plugin_cmds import lookup_pmr_rule
 from debputy.dh_migration.migrators import MIGRATORS
 from debputy.exceptions import (
     DebputyRuntimeError,
@@ -41,8 +40,10 @@ from debputy.exceptions import (
     PluginInitializationError,
     UnhandledOrUnexpectedErrorFromPluginError,
     SymlinkLoopError,
+    DebputyRuntimeErrorWithPreamble,
 )
 from debputy.highlevel_manifest import HighLevelManifest
+from debputy.manifest_parser.parser_doc import render_rule
 from debputy.package_build.assemble_deb import (
     assemble_debs,
 )
@@ -101,7 +102,7 @@ from debputy.util import (
     integrated_with_debhelper,
     PRINT_BUILD_SYSTEM_COMMAND,
     PRINT_COMMAND,
-    change_log_level,
+    manifest_format_doc,
 )
 
 
@@ -263,16 +264,11 @@ def _add_packages_args(parser: argparse.ArgumentParser) -> None:
     )
 
 
-def _build_subcommand_log_level(context: CommandContext) -> int:
-    parsed_args = context.parsed_args
-    log_level: Optional[int] = None
+def _build_subcommand_log_level(_context: CommandContext) -> int:
+    log_level: int = PRINT_BUILD_SYSTEM_COMMAND
     if os.environ.get("DH_VERBOSE", "") != "":
         log_level = PRINT_COMMAND
-    if parsed_args.debug_mode:
-        log_level = logging.INFO
-    if log_level is not None:
-        change_log_level(log_level)
-    return PRINT_BUILD_SYSTEM_COMMAND
+    return log_level
 
 
 internal_commands = ROOT_COMMAND.add_dispatching_subcommand(
@@ -283,7 +279,7 @@ internal_commands = ROOT_COMMAND.add_dispatching_subcommand(
 )
 tool_support_commands = ROOT_COMMAND.add_dispatching_subcommand(
     "tool-support",
-    help_description="Tool integration commands. These are intended to have stable output and behaviour",
+    help_description="Tool integration commands. These are intended to have stable output and behavior",
     dest="tool_subcommand",
     metavar="command",
 )
@@ -292,10 +288,19 @@ tool_support_commands = ROOT_COMMAND.add_dispatching_subcommand(
 def parse_args() -> argparse.Namespace:
     description = textwrap.dedent(
         """\
-    The `debputy` program is a manifest-based Debian packaging tool.
+    The `debputy` program is a Debian packaging tool.
 
-    It is used as a part of compiling a source package and transforming it into one or
-    more binary (.deb) packages.
+    It serves multiple roles in the Debian packaging stack:
+
+      1) It is a maintainer support tool that can help maintain packages by providing
+         editor support (LSP), batch linting and style/file formatting for Debian
+         packaging files. This is covers subcommands like `debputy lint`,
+         `debputy lsp server`, `debputy reformat`
+
+      2) It is a manifest-based Debian package builder aiming to replace existing
+         package helper tools such as `debhelper . In this role, `debputy` is used
+         as a part of compiling a source package and transforming it into one
+         or more binary (.deb) packages.
 
     If you are using a screen reader, consider exporting setting the environment variable
     OPTIMIZE_FOR_SCREEN_READER=1. This will remove some of the visual formatting and some
@@ -303,10 +308,18 @@ def parse_args() -> argparse.Namespace:
     """
     )
 
+    eplilog = textwrap.dedent(
+        """\
+
+    Bug tracker: https://salsa.debian.org/debian/debputy/-/issues
+    """
+    )
+
     parser: argparse.ArgumentParser = DebputyArgumentParser(
         description=description,
         formatter_class=argparse.RawDescriptionHelpFormatter,
         allow_abbrev=False,
+        epilog=eplilog,
         prog=program_name(),
     )
 
@@ -340,6 +353,9 @@ def parse_args() -> argparse.Namespace:
     if hasattr(parsed_args, "packages"):
         setattr(parsed_args, "packages", frozenset(parsed_args.packages))
 
+    if os.environ.get("DEBPUTY_DEBUG", "") == "1":
+        setattr(parsed_args, "debug_mode", True)
+
     return parsed_args
 
 
@@ -649,6 +665,64 @@ def _run_tests_for_plugin(context: CommandContext) -> None:
     _run_tests(context, tests, cwd="/")
 
 
+@internal_commands.register_subcommand(
+    "process-template",
+    help_description="[Internal command] Process a debputy template",
+    requested_plugins_only=True,
+    argparser=[
+        add_arg(
+            "template_file",
+            metavar="template-file",
+            help="The source template to be processed.",
+        ),
+        add_arg(
+            "output_file",
+            metavar="output-file",
+            help="Where to place the resulting file should be written",
+        ),
+    ],
+)
+def _process_template(context: CommandContext) -> None:
+    parsed_args = context.parsed_args
+    template_file = parsed_args.template_file
+    output_file = parsed_args.output_file
+
+    try:
+        from jinja2 import Environment, FileSystemLoader, Template
+    except ImportError:
+        _error("This sub-command requires a compatible version of python3-jinja2")
+
+    feature_set = context.load_plugins()
+
+    with open(template_file) as fd:
+        jinja_template = Template(
+            fd.read(),
+            autoescape=False,
+            block_start_string="<%",
+            block_end_string="%>",
+            variable_start_string="<<",
+            variable_end_string=">>",
+            comment_start_string="<#",
+            comment_end_string="#>",
+        )
+
+    def _render_pmr(rule_name: str, heading: str) -> str:
+        pmr_rule = lookup_pmr_rule(feature_set, rule_name)
+        return render_rule(
+            pmr_rule.rule_name,
+            pmr_rule.parser,
+            pmr_rule.plugin_metadata,
+            is_root_rule=pmr_rule.is_root_rule,
+            heading=heading,
+            include_ref_doc_link=False,
+            include_alt_format=False,
+            manifest_format_url="",
+        )
+
+    with open(output_file, "w") as fd:
+        fd.write(jinja_template.render(render_pmr=_render_pmr))
+
+
 @internal_commands.register_subcommand(
     "dpkg-build-driver-run-task",
     help_description="[Internal command] Perform a given Dpkg::BuildDriver task (Not stable API)",
@@ -916,6 +990,7 @@ def _annotate_debian_directory(context: CommandContext) -> None:
 
     result = scan_debian_dir(
         feature_set,
+        context.source_package(),
         binary_packages,
         context.debian_dir,
     )
@@ -1002,6 +1077,7 @@ def _migrate_from_dh(context: CommandContext) -> None:
         )
     )
     migrate_from_dh(
+        _output_styling(context.parsed_args, sys.stdout),
         manifest,
         acceptable_migration_issues,
         parsed_args.destructive,
@@ -1032,9 +1108,9 @@ def main() -> None:
         ROOT_COMMAND(cmd_arg)
     except PluginInitializationError as e:
         _error_w_stack_trace(
-            "Failed to load a plugin - full stack strace:",
+            "Failed to load a plugin - relevant stack trace (use --debug/DEBPUTY_DEBUG=1 to see the full one):",
             e.message,
-            e,
+            e.__cause__ if e.__cause__ else e,
             parsed_args.debug_mode,
             follow_warning=[
                 "Please consider filing a bug against the plugin in question"
@@ -1084,6 +1160,8 @@ def main() -> None:
                 "Re-raising original exception to show the full stack trace due to debug mode being active"
             )
             raise e
+        if isinstance(e, DebputyRuntimeErrorWithPreamble):
+            e.render_preamble()
         _error(e.message)
     except AssertionError as e:
         _error_w_stack_trace(
@@ -1109,7 +1187,7 @@ def main() -> None:
         )
     except Exception as e:
         _error_w_stack_trace(
-            "Unhandled exception (Re-run with --debug to see the raw stack trace)",
+            "Unhandled exception (Re-run with --debug/DEBPUTY_DEBUG=1 to see the raw stack trace)",
             str(e),
             e,
             parsed_args.debug_mode,
