@@ -12,8 +12,38 @@ from typing import (
     Iterator,
     Callable,
     cast,
+    TYPE_CHECKING,
 )
 
+from debputy.linting.lint_util import LintState, te_position_to_lsp
+from debputy.lsp.debputy_ls import DebputyLanguageServer
+from debputy.lsp.lsp_debian_control_reference_data import (
+    Deb822FileMetadata,
+    Deb822KnownField,
+    StanzaMetadata,
+    F,
+    S,
+    SUBSTVAR_RE,
+)
+from debputy.lsp.lsp_features import SEMANTIC_TOKEN_TYPES_IDS
+from debputy.lsp.quickfixes import propose_correct_text_quick_fix
+from debputy.lsp.text_util import (
+    trim_end_of_line_whitespace,
+    SemanticTokensState,
+)
+from debputy.lsp.vendoring._deb822_repro.locatable import (
+    START_POSITION,
+    Range as TERange,
+    Position as TEPosition,
+)
+from debputy.lsp.vendoring._deb822_repro.parsing import (
+    Deb822KeyValuePairElement,
+    Deb822ParagraphElement,
+    Deb822FileElement,
+    Deb822CommentElement,
+)
+from debputy.lsp.vendoring._deb822_repro.tokens import tokenize_deb822_file, Deb822Token
+from debputy.lsp.vendoring._deb822_repro.types import TokenOrElement
 from debputy.lsprotocol.types import (
     CompletionParams,
     CompletionList,
@@ -32,34 +62,14 @@ from debputy.lsprotocol.types import (
     MessageType,
     SemanticTokenTypes,
 )
-
-from debputy.linting.lint_util import LintState, te_position_to_lsp
-from debputy.lsp.debputy_ls import DebputyLanguageServer
-from debputy.lsp.lsp_debian_control_reference_data import (
-    Deb822FileMetadata,
-    Deb822KnownField,
-    StanzaMetadata,
-    F,
-    S,
-)
-from debputy.lsp.lsp_features import SEMANTIC_TOKEN_TYPES_IDS
-from debputy.lsp.text_util import (
-    trim_end_of_line_whitespace,
-    SemanticTokensState,
-)
-from debputy.lsp.vendoring._deb822_repro.locatable import (
-    START_POSITION,
-    Range as TERange,
-)
-from debputy.lsp.vendoring._deb822_repro.parsing import (
-    Deb822KeyValuePairElement,
-    Deb822ParagraphElement,
-    Deb822FileElement,
-)
-from debputy.lsp.vendoring._deb822_repro.tokens import tokenize_deb822_file, Deb822Token
-from debputy.lsp.vendoring._deb822_repro.types import TokenOrElement
 from debputy.util import _info, _warn
 
+if TYPE_CHECKING:
+    import lsprotocol.types as types
+else:
+    import debputy.lsprotocol.types as types
+
+
 try:
     from pygls.server import LanguageServer
     from pygls.workspace import TextDocument
@@ -472,6 +482,8 @@ class Deb822SemanticTokensState(SemanticTokensState):
         "keyword_token_code",
         "known_value_token_code",
         "comment_token_code",
+        "substvars_token_code",
+        "allow_overlapping_tokens",
     )
 
     def __init__(
@@ -484,15 +496,95 @@ class Deb822SemanticTokensState(SemanticTokensState):
         keyword_token_code: int,
         known_value_token_code: int,
         comment_token_code: int,
+        substvars_token_code: int,
     ) -> None:
         super().__init__(ls, doc, lines, tokens)
         self.file_metadata = file_metadata
         self.keyword_token_code = keyword_token_code
         self.known_value_token_code = known_value_token_code
         self.comment_token_code = comment_token_code
+        self.substvars_token_code = substvars_token_code
 
 
-def _deb822_paragraph_semantic_tokens_full(
+def _emit_tokens_for_comment_element(
+    sem_token_state: Deb822SemanticTokensState,
+    comment_element: Deb822CommentElement,
+    comment_start_line: int,
+    comment_token_code: int,
+) -> None:
+    for comment_line_no, comment_token in enumerate(
+        comment_element.iter_parts(),
+        start=comment_start_line,
+    ):
+        assert comment_token.is_comment
+        assert isinstance(comment_token, Deb822Token)
+        sem_token_state.emit_token(
+            Position(comment_line_no, 0),
+            len(comment_token.text.rstrip()),
+            comment_token_code,
+        )
+
+
+async def scan_for_syntax_errors_and_token_level_diagnostics(
+    deb822_file: Deb822FileElement,
+    lint_state: LintState,
+) -> int:
+    first_error = len(lint_state.lines) + 1
+    spell_checker = lint_state.spellchecker()
+
+    async for (
+        token,
+        start_line,
+        start_offset,
+        end_line,
+        end_offset,
+    ) in lint_state.slow_iter(deb822_token_iter(deb822_file.iter_tokens())):
+        if token.is_error:
+            first_error = min(first_error, start_line)
+            token_range = TERange(
+                TEPosition(
+                    start_line,
+                    start_offset,
+                ),
+                TEPosition(
+                    end_line,
+                    end_offset,
+                ),
+            )
+            lint_state.emit_diagnostic(
+                token_range,
+                "Syntax error",
+                "error",
+                "debputy",
+            )
+        elif token.is_comment:
+            for word, col_pos, end_col_pos in spell_checker.iter_words(token.text):
+                corrections = spell_checker.provide_corrections_for(word)
+                if not corrections:
+                    continue
+                word_range = TERange.between(
+                    TEPosition(
+                        start_line,
+                        col_pos,
+                    ),
+                    TEPosition(
+                        start_line,
+                        end_col_pos,
+                    ),
+                )
+                lint_state.emit_diagnostic(
+                    word_range,
+                    f'Spelling "{word}"',
+                    "spelling",
+                    "debputy",
+                    quickfixes=[propose_correct_text_quick_fix(c) for c in corrections],
+                    enable_non_interactive_auto_fix=False,
+                )
+    return first_error
+
+
+async def _deb822_paragraph_semantic_tokens_full(
+    ls: "DebputyLanguageServer",
     sem_token_state: Deb822SemanticTokensState,
     stanza: Deb822ParagraphElement,
     stanza_idx: int,
@@ -507,7 +599,9 @@ def _deb822_paragraph_semantic_tokens_full(
         stanza,
         stanza_idx=stanza_idx,
     )
-    for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
+    async for kvpair in ls.slow_iter(
+        stanza.iter_parts_of_type(Deb822KeyValuePairElement), yield_every=25
+    ):
         kvpair_position = kvpair.position_in_parent().relative_to(stanza_position)
         field_start = kvpair.field_token.position_in_parent().relative_to(
             kvpair_position
@@ -515,17 +609,13 @@ def _deb822_paragraph_semantic_tokens_full(
         comment = kvpair.comment_element
         if comment:
             comment_start_line = field_start.line_position - len(comment)
-            for comment_line_no, comment_token in enumerate(
-                comment.iter_parts(),
-                start=comment_start_line,
-            ):
-                assert comment_token.is_comment
-                assert isinstance(comment_token, Deb822Token)
-                sem_token_state.emit_token(
-                    Position(comment_line_no, 0),
-                    len(comment_token.text.rstrip()),
-                    comment_token_code,
-                )
+            _emit_tokens_for_comment_element(
+                sem_token_state,
+                comment,
+                comment_start_line,
+                comment_token_code,
+            )
+
         field_size = doc.position_codec.client_num_units(kvpair.field_name)
 
         sem_token_state.emit_token(
@@ -540,15 +630,37 @@ def _deb822_paragraph_semantic_tokens_full(
                 continue
             known_values: Container[str] = known_field.known_values or frozenset()
             interpretation = known_field.field_value_class.interpreter()
+            field_disallows_substvars = (
+                known_field.is_substvars_disabled_even_if_allowed_by_stanza
+            )
+            allow_substvars = (
+                stanza_metadata.is_substvars_allowed_in_stanza
+                and not field_disallows_substvars
+            )
         else:
             known_values = frozenset()
             interpretation = None
+            allow_substvars = stanza_metadata.is_substvars_allowed_in_stanza
 
         value_element_pos = kvpair.value_element.position_in_parent().relative_to(
             kvpair_position
         )
         if interpretation is None:
-            # TODO: Emit tokens for value comments of unknown fields.
+            for value_line in kvpair.value_element.value_lines:
+                comment_element = value_line.comment_element
+                if comment_element:
+                    assert comment_element.position_in_parent().line_position == 0
+                    comment_start_line = (
+                        value_line.position_in_parent()
+                        .relative_to(value_element_pos)
+                        .line_position
+                    )
+                    _emit_tokens_for_comment_element(
+                        sem_token_state,
+                        comment_element,
+                        comment_start_line,
+                        comment_token_code,
+                    )
             continue
         else:
             parts = kvpair.interpret_as(interpretation).iter_parts()
@@ -565,10 +677,17 @@ def _deb822_paragraph_semantic_tokens_full(
                 value = value.rstrip()
             elif value in known_values:
                 token_type = known_value_token_code
+            elif allow_substvars and "${" in value:
+                _process_value_with_substvars(
+                    sem_token_state,
+                    value,
+                    value_range_te,
+                    None,
+                )
+                continue
             else:
                 continue
             value_len = doc.position_codec.client_num_units(value)
-
             sem_token_state.emit_token(
                 te_position_to_lsp(value_range_te.start_pos),
                 value_len,
@@ -576,6 +695,70 @@ def _deb822_paragraph_semantic_tokens_full(
             )
 
 
+def _split_into_substvars(
+    value: str,
+    base_token_type: Optional[int],
+    substvar_token_type: int,
+) -> Iterable[Tuple[str, Optional[int]]]:
+
+    i = 0
+    next_search = i
+    full_value_len = len(value)
+    while i < full_value_len:
+        try:
+            subst_var_start = value.index("${", next_search)
+            subst_var_end = value.index("}", subst_var_start + 2)
+        except ValueError:
+            token = value[i:full_value_len]
+            if token:
+                yield token, base_token_type
+            return
+
+        subst_var_end += 1
+        subst_var = value[subst_var_start:subst_var_end]
+        if subst_var != "${}" and not SUBSTVAR_RE.match(subst_var):
+            subst_var = None
+
+        if subst_var is None:
+            next_search = subst_var_end
+            continue
+
+        token = value[i:subst_var_start]
+        if token:
+            yield token, base_token_type
+        yield subst_var, substvar_token_type
+        i = subst_var_end
+        next_search = i
+
+
+def _process_value_with_substvars(
+    sem_token_state: Deb822SemanticTokensState,
+    value: str,
+    value_range_te: "TERange",
+    base_token_type: Optional[int],
+) -> None:
+    pos_codec = sem_token_state.doc.position_codec
+
+    # TODO: Support overlapping tokens if the editor does.
+
+    line = value_range_te.start_pos.line_position
+    token_pos = value_range_te.start_pos.cursor_position
+    substvar_token_code = sem_token_state.substvars_token_code
+    for token, token_type in _split_into_substvars(
+        value,
+        base_token_type,
+        substvar_token_code,
+    ):
+        token_len = len(token)
+        if token_type is not None:
+            sem_token_state.emit_token(
+                types.Position(line, token_pos),
+                pos_codec.client_num_units(token),
+                token_type,
+            )
+        token_pos += token_len
+
+
 def deb822_format_file(
     lint_state: LintState,
     file_metadata: Deb822FileMetadata[Any, Any],
@@ -602,7 +785,7 @@ def deb822_format_file(
     )
 
 
-def deb822_semantic_tokens_full(
+async def deb822_semantic_tokens_full(
     ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
     file_metadata: Deb822FileMetadata[Any, Any],
@@ -629,11 +812,12 @@ def deb822_semantic_tokens_full(
         SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Keyword],
         SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.EnumMember],
         comment_token_code,
+        SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Macro],
     )
 
     stanza_idx = 0
 
-    for part in deb822_file.iter_parts():
+    async for part in ls.slow_iter(deb822_file.iter_parts(), yield_every=20):
         if part.is_comment:
             pos = part.position_in_file()
             sem_token_state.emit_token(
@@ -643,7 +827,8 @@ def deb822_semantic_tokens_full(
                 comment_token_code,
             )
         elif isinstance(part, Deb822ParagraphElement):
-            _deb822_paragraph_semantic_tokens_full(
+            await _deb822_paragraph_semantic_tokens_full(
+                ls,
                 sem_token_state,
                 part,
                 stanza_idx,
